/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.csv;

import java.util.Iterator;
import java.util.logging.Level;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.util.StringHelper;
import org.objectweb.petals.common.util.XMLUtil;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.se.AbstractServiceEngine;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.engine.csv.exception.CsvException;
import org.w3c.dom.Node;

/**
 * CSV service engine.
 * 
 * @version $Rev: $ $Date: $
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 */
public class CsvEngine extends AbstractServiceEngine {

    public static final String CSV_TO_XML_OPERATION = "csvToXml";
    public static final String XML_TO_CSV_OPERATION = "xmlToCsv";

    public static final String CSV_INPUT_MODE_CONTENT = "content";
    public static final String CSV_INPUT_MODE_ATTACHMENT = "attachment";
    public static final String CSV_INPUT_MODE_GUESS = "guess";
    public static final String CSV_INPUT_MODE_PROP = "csvInputMode";
    
    private static final String CONTENT_TAG = "content";
    
    private Csv csv = null;

    @Override
    public void start() throws JBIException {
        super.start();
        csv = new CsvImpl();
    }
    
    @Override
    public void stop() throws JBIException {
        super.stop();
        csv = null;
    }
    
    @Override
    protected boolean onExchange(MessageExchangeWrapper exchange, Extensions extensions)
            throws HandlingException {
        // Get Operation name
        QName operation = exchange.getOperation();

        // Get in Message content
        NormalizedMessage in = exchange.getInMessage();

        if (CSV_TO_XML_OPERATION.equals(operation.getLocalPart())) {
            try {
                // getting the content to be transformed
                String csvContent = getCsvInputContent(in, extensions);
                if (StringHelper.isNullOrEmpty(csvContent)) {
                    getLogger().log(Level.WARNING, "Empty message content");
                    // this will be automaticly added to a Fault if the message
                    throw new CsvException("operation : message empty");
                }

                // actual CSV to XML transformation
                Node xml = csv.transformCsvToXml(csvContent, extensions);
                String xmlString = XMLUtil.createStringFromDOMNode(xml, true);
                
                // Wraps the resulting XML in a <content> tag and returns it
                Source source = SourceHelper.createContentSource(xmlString);
                exchange.setOutMessageContent(source);
            } catch (MessagingException e) {
                throw new HandlingException("Error transforming message", e);
            } catch (Exception e) {
                String msg = "Error while transforming csv to XML : " + e.getMessage();
                getLogger().log(Level.SEVERE, msg, e);
                throw new HandlingException(msg, e);
            }

        } else if (XML_TO_CSV_OPERATION.equals(operation.getLocalPart())) {
            try {
                // getting the content to be transformed
                Node documentNode = getXmlInputDocument(in, extensions);
                Node xmlTableNode = XMLUtil.getFirstChild(documentNode);

                // actual XML to CSV transformation
                String csvData = csv.transformXmlToCsv(xmlTableNode, extensions);

                // Wraps the resulting CSV in a <content> tag and returns it
                Source source = SourceHelper.createContentSource(csvData);
                exchange.setOutMessageContent(source);
            } catch (Exception e) {
                String msg = "Error while transforming XML to CSV : " + e.getMessage();
                getLogger().log(Level.SEVERE, msg, e);
                throw new HandlingException(msg, e);
            }

        } else {
            throw new HandlingException("Unknown operation " + operation);
        }

        return true;
    }


    /**
     * Returns the actual CSV content from the given incoming message according
     * to the processor's settings
     * 
     * @param in the incoming message
     * @param extensions the service extensions
     * @return CSV content from the given incoming message
     * @throws Exception
     */
    protected String getCsvInputContent(NormalizedMessage in, Extensions extensions) throws Exception {
        String contentString;
        if (isModeAttachment(in, extensions)) {
            // getting first attachment
            String firstAttachmentName = null;
            for (Iterator nameIt = in.getAttachmentNames().iterator(); nameIt.hasNext();) {
                firstAttachmentName = (String) nameIt.next();
                break;
            }
            if (firstAttachmentName == null) {
                return null; // no content
            }
            DataHandler firstAttachment = in.getAttachment(firstAttachmentName);
            contentString = IOUtils.toString(firstAttachment.getInputStream());
            // NB. XML encoding of the ouput content is done during transformation

        } else {
            // inlined content mode, inside a <content> tag
            Source content = in.getContent();
            contentString = createStringFromContent(content);
            // NB. since it was already XML content, the values are already XML encoded
        }
        return contentString;
    }

    /**
     * Returns whether the CSV is to be looked for in the
     * message's first attachment instead of in its content
     * @param in the incoming message
     * @param extensions the service extensions
     * @return
     */
    protected boolean isModeAttachment(NormalizedMessage in, Extensions extensions) {
        String csvInputMode = PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, CSV_INPUT_MODE_PROP);
        if (CSV_INPUT_MODE_ATTACHMENT.equals(csvInputMode)) {
            return true;
        } else if (CSV_INPUT_MODE_CONTENT.equals(csvInputMode)) {
            return false;
        } else { // guessing mode
            return !in.getAttachmentNames().isEmpty();
        }
    }
    
    /**
     * Unwrapps the given content source : transforms the source
     * to a String and removes the (required) encapsulating &lt;content&gt;
     * tag.
     * @param content
     * @return content without &lt;content&gt; tag
     * @throws HandlingException
     */
    protected static String createStringFromContent(Source content) throws Exception {
        Node contentNode = XMLUtil.getFirstChild(createDOMNodeFromSource(content));
        // Checking that node element tag is "content"
        if (!CONTENT_TAG.equals(contentNode.getLocalName())) {
            throw new HandlingException("Expected <content> encapsulated content "
                + "but found " + XMLUtil.createStringFromDOMNode(contentNode));
        }
        // Usual case : the original content has been encapsulated as XML
        String contentString = contentNode.getTextContent();
        return contentString;
    }


    /**
     * Returns the actual XML content from the given incoming message
     * according to the processor's settings
     * @param in the incoming message
     * @param extensions the service extensions
     * @return
     * @throws HandlingException
     */
    protected Node getXmlInputDocument(NormalizedMessage in, Extensions extensions)
            throws HandlingException {
        try {
            if (isModeAttachment(in, extensions)) {
                // getting first attachment
                String firstAttachmentName = null;
                for (Iterator nameIt = in.getAttachmentNames().iterator(); nameIt.hasNext();) {
                    firstAttachmentName = (String) nameIt.next();
                    break;
                }
                if (firstAttachmentName == null) {
                    String msg = "Error while getting XML source content : "
                        + "inputMode is \"attachment\" but there is none given";
                    getLogger().log(Level.SEVERE, msg);
                    throw new CsvException(msg);
                }
                DataHandler firstAttachment = in.getAttachment(firstAttachmentName);
                String contentString = IOUtils.toString(firstAttachment.getInputStream());
                return XMLUtil.createDocumentFromString(contentString);

            } else {
                Source xmlTableSource = in.getContent();
                return createDOMNodeFromSource(xmlTableSource);
            }
        } catch (Exception e) {
            String msg = "Error while getting XML source content : " + e.getMessage();
            getLogger().log(Level.SEVERE, msg, e);
            throw new HandlingException(msg, e);
        }
    }

    /**
     * Create a DOM Node result from an XML Source
     * 
     * @param source the XML Source
     * @return a DOM representation of the XML Document
     * @throws TransformerException
     */
    public static Node createDOMNodeFromSource(Source source)
            throws TransformerException {
        DOMResult resultDom = new DOMResult();
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer;
        transformer = tFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
        transformer.transform(source, resultDom);
        Node resultNode = resultDom.getNode();
        return resultNode;
    }
}
