/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.csv;

import java.io.StringReader;

import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.util.XMLUtil;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.engine.csv.exception.CsvException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import com.csvreader.CsvReader;

/**
 * <p>This processor implements to operations : csvToXml and xmlToCsv.</p>
 * 
 * <p>
 * The csvToXml operation :<br/>
 * This processor uses Java CSV's CsvReader to transform either
 * the &lt;content&gt;-encapsulated content or the single attachment of
 * the message from any CSV format to XML.
 * </p>
 * 
 * <p>It is configured through the keyvalue extension in jbi.xml given
 * by the CSV SU handler. Here are the possible nodes :</p>
 * <ul>
 *     <li>csvInputMode : may be "content", "attachment" or "guess".
 * Default is "guess", meaning that if there is any attachment this
 * will be the chosen mode, and if not, it will be in "content" mode.
 * Not that this works for both ways (xmlToCsv and csvToXml)</li>
 *     <li>customColumnDelimiterChar : default is the comma ","</li>
 *     <li>useCustomRecordDelimiter : whether to use a custom one
 * instead of the default combination of unix & windows line breaks.
 * SO beware, using it lessens the original capabilities of the
 * parser and it is therefore not advised.</li>
 *     <li>customRecordDelimiterChar : custom line delimiter character.
 * See the "useCustomRecordDelimiter" property to activate this feature.</li>
 *     <li>useDoubleEscapeMode : whether to use doubled backslash to
 * trigger escaping instead of a single one.</li>
 *     <li>useComment : whether to parse and make comments available.
 * Default is false.</li>
 *     <li>commentChar : the character used to starts comments.
 * Default is sharp "#".</li>
 * </ul>
 * 
 * <p>
 * The xmlToCsv operation :<br/>
 * It views a two level-deep XML tree as tabular data and converts it
 * to CSV using the values of csvInputMode, customColumnDelimiterChar, 
 * useCustomRecordDelimiter and customRecordDelimiterChar with the same
 * meaning as above. The CSV is then returned in the response inside a
 * &lt;content&gt; top tag.
 * </p>
 * 
 * @version $Rev: $ $Date: $
 * @author Marc Dutoo - Open Wide
 * @author Guillaume Decarnin - Open Wide
 */
public class CsvImpl implements Csv {

    public static final String CUSTOM_COLUMN_DELIMITER_CHAR_PROP = "customColumnDelimiterChar";
    public static final String USE_CUSTOM_RECORD_DELIMITER_PROP = "useCustomRecordDelimiter";
    public static final String CUSTOM_RECORD_DELIMITER_CHAR_PROP = "customRecordDelimiterChar";
    public static final String USE_DOUBLE_ESCAPE_MODE_PROP = "useDoubleEscapeMode";
    public static final String USE_COMMENT_PROP = "useComment";
    public static final String COMMENT_CHAR_PROP = "commentChar";
    
    // The following constants describe the format of the XML output.
    // They could be configurable in properties.
    // The tag to be used to encapsulate CSV content in content input mode
    private static final String DATA_TAG = "data";
    private static final String ROW_TAG = "row";
    private static final String CSVXML_NAMESPACE_URI = "http://petals.objectweb.org/components/csv/csvxml";
    private static final String CELL_TAG = "cell";

    /**
     * Transforms the given CSV content input in XML and returns it
     * @param csvIn
     * @param extensions used to configure CSV parsing
     * @throws HandlingException
     */
    public Document transformCsvToXml(String csvIn, Extensions extensions)
            throws HandlingException {
        // creating a new CsvReader
        CsvReader csvReader = createCsvReader(csvIn, extensions);
        if (csvReader == null) {
            throw new CsvException("not a CSV file");
        }

        try {
            // now creating and outputting the transformed XML :
            // NB. using DOM so content will be XML encoded
            Document csvDoc = XMLUtil.createDocumentFromString("<" + DATA_TAG + " xmlns=\"" + CSVXML_NAMESPACE_URI + "\"/>");
            // NB. dataElement same as csvDoc
            Element dataElement = csvDoc.getDocumentElement();
            
            while (csvReader.readRecord()) {
                Element rowElement = csvDoc.createElementNS(CSVXML_NAMESPACE_URI, ROW_TAG);
                dataElement.appendChild(rowElement);
                int columnNb = csvReader.getColumnCount();
                
                for (int i = 0; i < columnNb; i++) {
                    String value = csvReader.get(i);
                    if (value == null) {
                        value = "";
                    }
                    
                    Element cellElement = csvDoc.createElementNS(CSVXML_NAMESPACE_URI, CELL_TAG);
                    rowElement.appendChild(cellElement);
                    // NB. we could encapsulating CSV values in CDATA so anything should be ok
                    //CDATASection valueData = csvDoc.createCDATASection(value);
                    Text valueData = csvDoc.createTextNode(value);
                    cellElement.appendChild(valueData);
                }
            }
            
            return csvDoc;
        } catch (Exception e) {
            throw new HandlingException("Error while parsing CSV and building output XML : "
                + e.getMessage());
        }
    }

    /**
     * Creates a new CsvReader configured according to the
     * component's properties, ready to parse CSV content.
     * @param csv the CSV content to be parser
     * @param extensions used to configure the CsvReader
     * @return a new, configured CsvReader
     * @throws HandlingException if creation or configuration error
     */
    protected CsvReader createCsvReader(String csv, Extensions extensions) throws HandlingException {
        try {
            char csvColumnDelimiterChar = extractValueFromKeyValueExtension(extensions, CUSTOM_COLUMN_DELIMITER_CHAR_PROP, ",").charAt(0);
            char csvCustomRecordDelimiterChar = extractValueFromKeyValueExtension(extensions, CUSTOM_RECORD_DELIMITER_CHAR_PROP, "\n").charAt(0);
            boolean useCustomRecordDelimiter = Boolean.parseBoolean(extractValueFromKeyValueExtension(extensions, USE_CUSTOM_RECORD_DELIMITER_PROP, "false"));
            
            boolean useDoubleEscapeMode = Boolean.parseBoolean(extractValueFromKeyValueExtension(extensions, USE_DOUBLE_ESCAPE_MODE_PROP, "true"));
            int escapeMode = useDoubleEscapeMode ? CsvReader.ESCAPE_MODE_DOUBLED : CsvReader.ESCAPE_MODE_BACKSLASH;
            
            boolean useComment = Boolean.parseBoolean(extractValueFromKeyValueExtension(extensions, USE_COMMENT_PROP, "false"));
            char commentChar = extractValueFromKeyValueExtension(extensions, COMMENT_CHAR_PROP, "#").charAt(0);
        
            // creating CsvReader for given input
            // csvCharset : Csv SE default is inited in TODO
            CsvReader csvReader = new CsvReader(new StringReader(csv));
            csvReader.setDelimiter(csvColumnDelimiterChar); // CsvReader default is ','
            if (useCustomRecordDelimiter) {
               csvReader.setRecordDelimiter(csvCustomRecordDelimiterChar);
            } // else CsvReader default is a combination of unix & windows cr & lf
            csvReader.setEscapeMode(escapeMode); // CsvReader default is CsvReader.ESCAPE_MODE_DOUBLED i.e. '\\'
            csvReader.setUseComments(useComment);
            csvReader.setComment(commentChar); // CsvReader default is '#'
            // NB. safetySwitch is activated and may not be disabled for now
            //csvReader.setMergeConsecutive(arg0) //TODO ???
            return csvReader;
        } catch (Exception e) {
            throw new HandlingException("Error setting CSV reader configuration. CSV reader properties are : " + extensions);
        }
    }
    
    /**
     * Transforms the given XML tabular data to CSV.
     * There is no requirement on the format (tags...) besides the fact that
     * they have to be a two level tree.
     * @param xml the XML tabular data to be written as CSV
     * @param extensions the service extensions
     * @return CSV String
     * @throws HandlingException if conversion error
     */
    public String transformXmlToCsv(Node xml, Extensions extensions) throws HandlingException {
        // getting CSV formatting properties
        char csvColumnDelimiterChar = extractValueFromKeyValueExtension(extensions, CUSTOM_COLUMN_DELIMITER_CHAR_PROP, ",").charAt(0);
        char csvCustomRecordDelimiterChar = extractValueFromKeyValueExtension(extensions, CUSTOM_RECORD_DELIMITER_CHAR_PROP, "\n").charAt(0);
        boolean useCustomRecordDelimiter = Boolean.parseBoolean(extractValueFromKeyValueExtension(extensions, USE_CUSTOM_RECORD_DELIMITER_PROP, "false"));
        String recordDelimiter = (useCustomRecordDelimiter) ? String.valueOf(csvCustomRecordDelimiterChar) : "\r\n";
        
        StringBuilder csvBuf = new StringBuilder();
            
        NodeList rowNodes = xml.getChildNodes();
        int rowNodeNb = rowNodes.getLength();
        for (int i = 0; i < rowNodeNb; i++) {
            Node rowNode = rowNodes.item(i);
            if (!(rowNode instanceof Element)) {
                continue;
            }
            
            // now outputting the current row :
            NodeList columnNodes = rowNode.getChildNodes();
            int columnNodeNb = columnNodes.getLength();
            for (int j = 0; j < columnNodeNb; j++) {
                Node columnNode = columnNodes.item(j);
                if (!(columnNode instanceof Element)) {
                    continue;
                }
                
                // outputting cell value
                String cellValue = XMLUtil.getTextContent(columnNode);
                csvBuf.append(cellValue);
                csvBuf.append(csvColumnDelimiterChar);
            }
            if (columnNodeNb != 0) {
                // removing the last column delimiter
                csvBuf.deleteCharAt(csvBuf.length() - 1);
            }
            
            // appending end of row
            csvBuf.append(recordDelimiter);
        }
        
        if (rowNodeNb != 0) {
            // removing the last row delimiter
            csvBuf.deleteCharAt(csvBuf.length() - recordDelimiter.length());
        }
        
        return csvBuf.toString();
    }
    
    /**
     * Extract a value from the key value map extension.
     * The method returns the default value argument if the property is not found.
     * @param extensions the {@link Extensions} element containing Petals key value extension
     * @param key the targeted key
     * @param defaultValue a default value
     * @return the {@link String} value for the given key or null
     */
    private String extractValueFromKeyValueExtension(Extensions extensions, String key, String defaultValue) {
        if (PetalsExtensionsUtil.hasKeyValueExtension(extensions, key)) {
            return PetalsExtensionsUtil.extractValueFromKeyValueExtension(extensions, key);
        } else {
            return defaultValue;
        }
    }
}
